local _M = loadPrevious(...)
require "engine.class"
require "engine.Actor"
require "engine.Autolevel"
require "engine.interface.ActorInventory"
require "engine.interface.ActorTemporaryEffects"
require "mod.class.interface.ActorLife"
require "engine.interface.ActorProject"
require "engine.interface.ActorLevel"
require "engine.interface.ActorStats"
require "engine.interface.ActorTalents"
require "engine.interface.ActorResource"
require "engine.interface.BloodyDeath"
require "engine.interface.ActorFOV"
require "mod.class.interface.ActorPartyQuest"
require "mod.class.interface.Combat"
require "mod.class.interface.Archery"
require "mod.class.interface.ActorInscriptions"
local Faction = require "engine.Faction"
local Dialog = require "engine.ui.Dialog"
local Map = require "engine.Map"
local DamageType = require "engine.DamageType"

local onTakeHit = _M.onTakeHit
function _M:onTakeHit(value, src, death_note)
	src = src or {}
	if value <=0 then return onTakeHit(self, 0, src, death_note) end
	local tal
	-- update hate_baseline
	if self.knowTalent and self:knowTalent(self.T_HATE_POOL) then
		local t = self:getTalentFromId(self.T_HATE_POOL)
		t.updateBaseline(self, t)
	end
	if src and src.knowTalent and src:knowTalent(src.T_HATE_POOL) then
		local t = src:getTalentFromId(src.T_HATE_POOL)
		t.updateBaseline(src, t)
	end

	-- Un-daze
	if self:hasEffect(self.EFF_DAZED) and not self:attr("damage_dont_undaze") then
		self:removeEffect(self.EFF_DAZED)
	end

	if self:isTalentActive(self.T_SUSPENDED) then
		self:forceUseTalent(self.T_SUSPENDED, {ignore_energy=true})
	end

	-- Remove domination hex
	if self:hasEffect(self.EFF_DOMINATION_HEX) and src and src == self:hasEffect(self.EFF_DOMINATION_HEX).src then
		self:removeEffect(self.EFF_DOMINATION_HEX)
	end

	if self:attr("invulnerable") then
		return onTakeHit(self, 0, src, death_note)
	end
	if self:attr("invulnerable_others") and src ~= self then
		return onTakeHit(self, 0, src, death_note)
	end

	-- verbose addon start
	-- we record here the initial value of dmg, so we can see the total reduction of damage after all the steps in this function
	-- we only do this for the player, not npcs, summons etc
	local jx_value_start = value
	local jx_value_end = value
	if self == game.player then
		print("$$$ Initial value of this damage is: ", jx_value_start)
	end
	-- verbose addon end

	if self:attr("cancel_damage_chance") and rng.percent(self.cancel_damage_chance) then
		if self:knowTalent(self.T_TEMPO) then
			local t = self:getTalentFromId(self.T_TEMPO)
			t.do_tempo(self, t)
		end
		-- verbose addon start
		-- these can be checked sequentially, because none of them are compatible with each other. they require, respectively, two steamsaws, two mindstars, 1 2Hander, archery
		-- I do not know check self:isTalentActive because if cancel_damage_chance has a value the talent has to be active
		--print("$$$ cancel_damage_chance of ", self.cancel_damage_chance," negated damage of value: ", value)
		if self:knowTalent(self.T_TEMPEST_OF_METAL) and self == game.player then
			print("$$$ Tempest of Metal has negated all damage, of value: ", value, "(", self.cancel_damage_chance, "%)")
		end
		if self:knowTalent(self.T_LEAVES_TIDE) and self == game.player then
			print("$$$ Leaves Tide has negated all damage, of value: ", value, "(", self.cancel_damage_chance, "%)")
		end
		if self:knowTalent(self.T_FLASH_OF_THE_BLADE) and self == game.player then
			print("$$$ Flash of the Blade has all avoided damage, of value: ", value, "(", self.cancel_damage_chance, "%)")
		end
		if self:knowTalent(self.T_CONCEALMENT) and self == game.player then
			print("$$$ Concealment has avoided all damage, of value: ", value, "(", self.cancel_damage_chance, "%)")
		end
		-- verbose addon end
		return onTakeHit(self, 0, src, death_note)
	end

	if self:attr("phase_shift") and rng.percent(self:attr("phase_shift") * 100) and not self.turn_procs.phase_shift then
		self.turn_procs.phase_shift = true
		local nx, ny = util.findFreeGrid(self.x, self.y, 1, true, {[Map.ACTOR]=true})
		if nx then
			local ox, oy = self.x, self.y
			self:move(nx, ny, true)
			game.level.map:particleEmitter(ox, oy, math.max(math.abs(nx-ox), math.abs(ny-oy)), "lightning", {tx=nx-ox, ty=ny-oy})
			game:delayedLogDamage(src or {}, self, 0, ("#STEEL_BLUE#(%d shifted)#LAST#"):format(value), nil)
			-- verbose addon start
			if self == game.player then
				print("$$$ phase_shift has removed all damage! elemental surge, chonomancy, fast as lightning, or lucky day is the cause")	
			end
			-- verbose addon end
			return onTakeHit(self, 0, src, death_note)
		end
	end

	if self:attr("retribution") then
	-- Absorb damage into the retribution
		-- verbose addon start
		local oldval = value
		-- verbose addone end
		local absorb = math.min(value/2, self.retribution_absorb)
		game:delayedLogDamage(src, self, 0, ("#SLATE#(%d absorbed)#LAST#"):format(absorb), false)
		if absorb < self.retribution_absorb then
			self.retribution_absorb = self.retribution_absorb - absorb
			value = value - absorb
		else
			value = value - self.retribution_absorb
			self.retribution_absorb = 0
			local dam = self.retribution_strike

			-- Deactivate without loosing energy
			self:forceUseTalent(self.T_RETRIBUTION, {ignore_energy=true, ignore_cd=true})
			self:startTalentCooldown(self.T_RETRIBUTION)

			-- Explode!
			game.logSeen(self, "%s unleashes the stored damage in retribution!", self.name:capitalize())
			local tg = {type="ball", range=0, radius=self:getTalentRange(self:getTalentFromId(self.T_RETRIBUTION)), selffire=false, talent=t}
			local grids = self:project(tg, self.x, self.y, DamageType.LIGHT, dam)
			game.level.map:particleEmitter(self.x, self.y, tg.radius, "sunburst", {radius=tg.radius, grids=grids, tx=self.x, ty=self.y})
		end
		-- verbose addon start
		if self == game.player then
			print("$$$ Retribution absorbed damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end
	end

	--Special Flag (for older versions of Terrasca)
	if value > 0 and self:attr("speed_resist") then
		value = value * (util.bound(self.global_speed * self.movement_speed, 0.3, 1))
	end

	-- General percent damage reduction
	if self:attr("incoming_reduce") then
		local oldval = value
		value = value * (100-self:attr("incoming_reduce")) / 100
		print("[onTakeHit] After Trained Reactions effect reduction ", value)
		-- verbose addon start
		if self:knowTalent(self.T_SNIPE) and self == game.player then
			print("$$$ Snipe has reduced damage by value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		if self:knowTalent(self.T_SKIRMISHER_TRAINED_REACTIONS) and self == game.player then
			print("$$$ Skirmisher Trained Reactions has reduced damage by value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end
	end

	if self:knowTalent(self.T_SKIRMISHER_TRAINED_REACTIONS) then
		local t = self:getTalentFromId(self.T_SKIRMISHER_TRAINED_REACTIONS)
		if self:isTalentActive(t.id) then
			value = t.onHit(self, t, value)
			-- print("[onTakeHit] After Trained Reactions life% trigger ", value) changed to make more accurate by jenx, verbose
			print("[onTakeHit] After Skirmisher Trained Reactions life% trigger ", value)
		end
	end

	if value > 0 and self:knowTalent(self.T_MITOSIS) and self:isTalentActive(self.T_MITOSIS) then
		local t = self:getTalentFromId(self.T_MITOSIS)
		local chance = t.getChance(self, t)
		local perc = math.min(1, 3 * value / self.life)
		if rng.percent(chance * perc) then
			t.spawn(self, t, value * 2)
		end

		local acts = {}
		if game.party:hasMember(self) then
			for act, def in pairs(game.party.members) do
				if act.summoner and act.summoner == self and act.wild_gift_summon and act.bloated_ooze then acts[#acts+1] = act end
			end
		else
			for _, act in pairs(game.level.entities) do
				if act.summoner and act.summoner == self and act.wild_gift_summon and act.bloated_ooze then acts[#acts+1] = act end
			end
		end
		if #acts > 0 then
			local oldval = value
			game:delayedLogMessage(self, nil, "mitosis_damage", "#DARK_GREEN##Source# shares damage with %s oozes!", string.his_her(self))
			value = value / (#acts+1)
			for _, act in ipairs(acts) do
				act:takeHit(value, src)
			end
			-- verbose addon start
			if self == game.player then
				print("$$$ Mitosis has absorbed damage of value: ", oldval - value, ", with this much damage remaining: ", value)
			end
			-- verbose addon end
		end
	end

	if value > 0 and self:attr("displacement_shield") then
		-- Absorb damage into the displacement shield
		if rng.percent(self.displacement_shield_chance) then
			game:delayedLogMessage(self, src,  "displacement_shield"..(self.displacement_shield_target.uid or ""), "#CRIMSON##Source# teleports some damage to #Target#!")
			local displaced = math.min(value, self.displacement_shield)
			self.displacement_shield_target:takeHit(displaced, src)
			game:delayedLogDamage(src, self, 0, ("#CRIMSON#(%d teleported)#LAST#"):format(displaced), false)
			game:delayedLogDamage(src, self.displacement_shield_target, displaced, ("#CRIMSON#%d teleported#LAST#"):format(displaced), false)
			if self.displacement_shield and displaced < self.displacement_shield then
				self.displacement_shield = self.displacement_shield - displaced
				
				-- verbose addon start
				if self == game.player then
					print("$$$ Displacement Shield has absorbed ALL damage, of value: ", value)
				end
				-- verbose addon start

				value = 0
			else
				self:removeEffect(self.EFF_DISPLACEMENT_SHIELD)

				-- verbose addon start
				if self == game.player then
					print("$$$ Displacement Shield has absorbed damage of value: ", displaced, ", with this much damage remaining: ", value - displaced)
				end
				-- verbose addon start

				value = value - displaced
			end
		end
	end

	if value > 0 and self:attr("time_shield") then
		-- Absorb damage into the time shield
		self.time_shield_absorb = self.time_shield_absorb or 0
		game:delayedLogDamage(src, self, 0, ("#STEEL_BLUE#(%d to time)#LAST#"):format(math.min(value, self.time_shield_absorb)), false)
		if value < self.time_shield_absorb then
			self.time_shield_absorb = self.time_shield_absorb - value
				
			-- verbose addon start
			if self == game.player then
				print("$$$ Time Shield has absorbed ALL damage, of value: ", value)
			end
			-- verbose addon start

			value = 0
		else

			-- verbose addon start
			if self == game.player then
				print("$$$ Time Shield has absorbed damage of value: ", self.time_shield_absorb, ", with this much damage remaining: ", value - self.time_shield_absorb)
			end
			-- verbose addon start

			value = value - self.time_shield_absorb
			self.time_shield_absorb = 0
		end

		-- If we are at the end of the capacity, release the time shield damage
		if self.time_shield_absorb <= 0 then
			game.logPlayer(self, "Your time shield crumbles under the damage!")
			self:removeEffect(self.EFF_TIME_SHIELD)
		end
	end

	if value > 0 and self:attr("damage_shield") then
		-- Phased attack?
		local adjusted_value = value
		if src and src.attr and src:attr("damage_shield_penetrate") then
			adjusted_value = value * (1 - (util.bound(src.damage_shield_penetrate, 0, 100) / 100))
		end
		-- Shield Reflect?
		local reflection, reflect_damage = 0
		if self:attr("damage_shield_reflect") then
			reflection = self:attr("damage_shield_reflect")/100
		end
		-- Absorb damage into the shield
		self.damage_shield_absorb = self.damage_shield_absorb or 0

		-- verbose addon start DEBUG
		local oldshield = self.damage_shield_absorb
		print("$$$ Damage Shield DEBUG. INIT: Value:", value, " adjusted_value:", adjusted_value, " shield value: ", self.damage_shield_absorb, " penetration:", src.damage_shield_penetrate or 0)
		-- verbose addon end

		if adjusted_value <= self.damage_shield_absorb then
			self.damage_shield_absorb = self.damage_shield_absorb - adjusted_value
			if reflection > 0 then reflect_damage = adjusted_value end
			value = value - adjusted_value

			print("$$$ Damage Shield DEBUG. adjusted<=shield: Value:", value, " adjusted_value:", adjusted_value, " shield value: ", self.damage_shield_absorb, " penetration:", src.damage_shield_penetrate or 0)

		else
			if reflection > 0 then reflect_damage = self.damage_shield_absorb end
			value = adjusted_value - self.damage_shield_absorb
			adjusted_value = self.damage_shield_absorb
			self.damage_shield_absorb = 0

			print("$$$ Damage Shield DEBUG. adjusted>shield: Value:", value, " adjusted_value:", adjusted_value, " shield value: ", self.damage_shield_absorb, " penetration:", src.damage_shield_penetrate or 0)

		-- verbose addon start
		if self == game.player then
			print("$$$ Damage Shield has absorbed damage of value: ", oldshield - self.damage_shield_absorb, ", with this much damage remaining: ", value)
		end
		-- verbose addon start

		end
		game:delayedLogDamage(src, self, 0, ("#SLATE#(%d absorbed)#LAST#"):format(adjusted_value), false)
		if reflection and reflect_damage and reflection > 0 and reflect_damage > 0 and src.y and src.x and not src.dead then
			local a = game.level.map(src.x, src.y, Map.ACTOR)
			if a and self:reactionToward(a) < 0 then
				local reflected = reflect_damage * reflection
				a:takeHit(reflected, self)
				game:delayedLogDamage(self, src, reflected, ("#SLATE#%d reflected#LAST#"):format(reflected), false)
				game:delayedLogMessage(self, src, "reflection" ,"#CRIMSON##Source# reflects damage back to #Target#!#LAST#")
			end
		end
		-- If we are at the end of the capacity, release the time shield damage
		if not self.damage_shield_absorb or self.damage_shield_absorb <= 0 then
			game.logPlayer(self, "Your shield crumbles under the damage!")
			self:removeEffect(self.EFF_DAMAGE_SHIELD)
			self:removeEffect(self.EFF_PSI_DAMAGE_SHIELD)
		end
	end

	if value > 0 and self:attr("shadow_empathy") then
		-- Absorb damage into a random shadow
		local shadow = self:callTalent(self.T_SHADOW_EMPATHY, "getRandomShadow")
		if shadow then
			game:delayedLogMessage(self, src,  "displacement_shield"..(shadow.uid or ""), "#CRIMSON##Source# shares some damage with a shadow!")
			local displaced = math.min(value * self.shadow_empathy / 100, shadow.life)
			shadow:takeHit(displaced, src)
			game:delayedLogDamage(src, self, 0, ("#PINK#(%d linked)#LAST#"):format(displaced), false)
			game:delayedLogDamage(src, shadow, displaced, ("#PINK#%d linked#LAST#"):format(displaced), false)
			value = value - displaced
		
			-- verbose addon start
			if self == game.player then
				print("$$$ Shadow Empathy has absorbed damage of value: ", displaced, ", with this much damage remaining: ", value)
			end
			-- verbose addon start

		end
	end

	if value > 0 and self:isTalentActive(self.T_DISRUPTION_SHIELD) then
		local mana = math.max(0, self:getMaxMana() - self:getMana())
		local mana_val = value * self:attr("disruption_shield")
		local converted = math.min(value, mana / self:attr("disruption_shield"))

		local oldval = value -- verbose

		game:delayedLogMessage(self, nil,  "disruption_shield", "#LIGHT_BLUE##Source# converts damage to mana!")
		game:delayedLogDamage(src, self, 0, ("#LIGHT_BLUE#(%d converted)#LAST#"):format(converted), false)

		-- We have enough to absorb the full hit
		if mana_val <= mana then
			self:incMana(mana_val)
			self.disruption_shield_absorb = self.disruption_shield_absorb + value

			-- verbose addon start
			if self == game.player then
				print("$$$ Disruption Shield has absorbed ALL damage, of value: ", oldval)
			end
			-- verbose addon start

			return onTakeHit(self, 0, src, death_note)
		-- Or the shield collapses in a deadly arcane explosion
		else
			self:incMana(mana)
			self.disruption_shield_absorb = self.disruption_shield_absorb + mana / self:attr("disruption_shield")
			value = value - mana / self:attr("disruption_shield")

			-- verbose addon start
			if self == game.player then
				print("$$$ Disruption Shield has absorbed damage of value: ", oldval - value, ", with this much damage remaining: ", value)
			end
			-- verbose addon start

			local dam = self.disruption_shield_absorb

			-- Deactivate without loosing energy
			self:forceUseTalent(self.T_DISRUPTION_SHIELD, {ignore_energy=true})

			-- Explode!
			local t = self:getTalentFromId(self.T_DISRUPTION_SHIELD)
			t.explode(self, t, dam)
		end
	end

	if value > 0 and self:isTalentActive(self.T_BONE_SHIELD) then
		local oldval = value
		local t = self:getTalentFromId(self.T_BONE_SHIELD)
		if t.absorb(self, t, self:isTalentActive(self.T_BONE_SHIELD)) then
			game:delayedLogDamage(src, self, 0, ("#SLATE#(%d to bones)#LAST#"):format(value), false)
			value = 0
		-- verbose addon start
		if self == game.player then
			print("$$$ Bone Shield has absorbed one lot of damage, of value: ", oldval)
		end
		-- verbose addon end
		end
	end

	if value <=0 then return onTakeHit(self, 0, src, death_note) end
	if self.knowTalent and (self:knowTalent(self.T_SEETHE) or self:knowTalent(self.T_GRIM_RESOLVE)) then
		if not self:hasEffect(self.EFF_CURSED_FORM) then
			self:setEffect(self.EFF_CURSED_FORM, 1, { increase=0 })
		end
		local eff = self:hasEffect(self.EFF_CURSED_FORM)
		self.tempeffect_def[self.EFF_CURSED_FORM].do_onTakeHit(self, eff, value)
	end

	if value > 0 then
		tal = self:isTalentActive(self.T_DEFLECTION)
		if tal then
			local oldval = value
			value = self:callTalent(self.T_DEFLECTION, "do_onTakeHit", tal, value)
			if value ~= oldval then
				game:delayedLogDamage(src, self, 0, ("#SLATE#(%d deflected)#LAST#"):format(oldval - value), false)

				-- verbose addon start
				if self == game.player then
					print("$$$ Deflection has deflected damage of value: ", oldval - value, ", with this much damage remaining: ", value)
				end
				-- verbose addon end

			end
		end
	end

	if value > 0 and self:hasEffect(self.EFF_RAMPAGE) then
		local eff = self:hasEffect(self.EFF_RAMPAGE)
		local oldval = value
		value = self.tempeffect_def[self.EFF_RAMPAGE].do_onTakeHit(self, eff, value)

		-- verbose addon start
		if self == game.player then
			print("$$$ Rampage has shrugged off damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end

	end

	if value > 0 and self:hasEffect(self.EFF_BECKONED) then
		local eff = self:hasEffect(self.EFF_BECKONED)
		value = self.tempeffect_def[self.EFF_BECKONED].do_onTakeHit(self, eff, value)
	end

	-- Achievements
	if not self.no_take_hit_achievements and src and src.resolveSource and src:resolveSource().player and value >= 600 then
		local rsrc = src:resolveSource()
		world:gainAchievement("SIZE_MATTERS", rsrc)
		if value >= 1500 then world:gainAchievement("DAMAGE_1500", rsrc) end
		if value >= 3000 then world:gainAchievement("DAMAGE_3000", rsrc) end
		if value >= 6000 then world:gainAchievement("DAMAGE_6000", rsrc) end
	end

	-- Frozen: absorb some damage into the iceblock
	if value > 0 and self:attr("encased_in_ice") then
		local eff = self:hasEffect(self.EFF_FROZEN)
		local absorb = 0.4
		local oldval = value --verbose
		if src and src.attr then
			absorb = absorb - absorb * (util.bound(src:attr("iceblock_pierce") or 0, 0, 100)) / 100
		end
		eff.hp = eff.hp - value * absorb
		game:delayedLogDamage(src or {}, self, 0, ("#STEEL_BLUE#(%d to ice)#LAST#"):format(value*absorb), nil)
		value = value * (1 - absorb)

		-- verbose addon start
		if self == game.player then
			print("$$$ Iceblock has shrugged off damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end

		if eff.hp < 0 and not eff.begone then
			game:onTickEnd(function() self:removeEffect(self.EFF_FROZEN) end)
			eff.begone = game.turn
		end
	end

	-- Mind save to reduce damage to zero
	if self:knowTalent(self.T_DISMISSAL) and value > 0 then
		local oldval = value -- verbose
		local t = self:getTalentFromId(self.T_DISMISSAL)
		value = t.doDismissalOnHit(self, value, src, t)

		-- verbose addon start
		if self == game.player then
			print("$$$ Dismissal has shrugged off damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end

	end

	if value <= 0 then return onTakeHit(self, 0, src, death_note) end
	-- Feedback pool: Stores damage as energy to use later
	if self:getMaxFeedback() > 0 and src ~= self and src ~= self.summoner then
		local ratio = self:callTalent(self.T_FEEDBACK_POOL, "getFeedbackRatio")
		local feedback_gain = value * ratio
		self:incFeedback(feedback_gain)
		-- Give feedback to summoner
		if self.summoner and self.summoner:getTalentLevel(self.summoner.T_OVER_MIND) >=1 and self.summoner:getMaxFeedback() > 0 then
			self.summoner:incFeedback(feedback_gain)
		end
		-- Trigger backlash retribution damage
		if src and src.turn_procs and self:knowTalent(self.T_BACKLASH) and not src.no_backlash_loops and not src.turn_procs.backlash then
			if src.y and src.x and not src.dead then
				local t = self:getTalentFromId(self.T_BACKLASH)
				t.doBacklash(self, src, feedback_gain, t)
				src.turn_procs.backlash = true
			end
		end
	end

	-- Resonance Field, must be called after Feedback gains
	if self:attr("resonance_field") then
		local oldval = value -- Verbose
		local absorb = math.min(value/2, self.resonance_field_absorb)
		game:delayedLogDamage(src, self, 0, ("#SLATE#(%d resonance)#LAST#"):format(absorb), false)
		if absorb < self.resonance_field_absorb then
			self.resonance_field_absorb = self.resonance_field_absorb - absorb
		else
			game.logPlayer(self, "Your resonance field crumbles under the damage!")
			self:removeEffect(self.EFF_RESONANCE_FIELD)
		end
		value = value - absorb

		-- verbose addon start
		if self == game.player then
			print("$$$ Resonance Field has shrugged off damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end

	end

	if self:hasEffect(self.EFF_CAUTERIZE) then
		
		-- verbose addon start
		if self == game.player then
			print("$$$ Cauterize has removed ALL damage, of value: ", value)
		end
		-- verbose addon end

		local eff = self:hasEffect(self.EFF_CAUTERIZE)
		if eff.invulnerable then
			eff.dam = eff.dam + value / 10
			return onTakeHit(self, 0, src, death_note)
		end
	end

	-- Reduce sleep durations
	if value > 0 and self:attr("sleep") then
		local effs = {}
		for eff_id, p in pairs(self.tmp) do
			local e = self.tempeffect_def[eff_id]
			if e.subtype.sleep then
				effs[#effs+1] = {"effect", eff_id}
			end
		end
		for i = 1, #effs do
			if #effs == 0 then break end
			local eff = rng.tableRemove(effs)
			if eff[1] == "effect" then
				local e = self:hasEffect(eff[2])
				-- Reduce the duration by 1 for every full incriment of the effect power
				-- We add a temp_power parameter to track total damage over multiple turns
				e.temp_power = (e.temp_power or e.power) - value
				while e.temp_power <= 0 do
					e.dur = e.dur - 1
					e.temp_power = e.temp_power + e.power
					if e.dur <=0 then
						game:onTickEnd(function() self:removeEffect(eff[2]) end) -- Happens on tick end so Night Terror can work properly
						break
					end
				end
			end
		end
	end

	-- Solipsism damage; set it up here but apply it after on_take_hit
	local damage_to_psi = 0
	if self:knowTalent(self.T_SOLIPSISM) and value > 0 and self:getPsi() > 0 then
		local t = self:getTalentFromId(self.T_SOLIPSISM)
		damage_to_psi = value * t.getConversionRatio(self, t)
	end

	-- Stoned ? SHATTER !
	if self:attr("stoned") and value >= self.max_life * 0.3 then
		-- Make the damage high enough to kill it
		value = self.max_life + 1
		game.logSeen(self, "%s shatters into pieces!", self.name:capitalize())
	end

	-- Adds hate
	if value > 0 and self:knowTalent(self.T_HATE_POOL) then
		local hateGain = 0
		local hateMessage

		if value / self.max_life >= 0.15 then
			-- you take a big hit..adds 2 + 2 for each 5% over 15%
			hateGain = hateGain + 2 + (((value / self.max_life) - 0.15) * 100 * 0.5)
			hateMessage = "#F53CBE#You fight through the pain!"
		end

		if value / self.max_life >= 0.05 and (self.life - value) / self.max_life < 0.25 then
			-- you take a hit with low health
			hateGain = hateGain + 4
			hateMessage = "#F53CBE#Your hatred grows even as your life fades!"
		end

		if hateGain >= 1 then
			self:incHate(hateGain)
			if hateMessage then
				game.logPlayer(self, hateMessage.." (+%d hate)", hateGain)
			end
		end
	end

	if value > 0 and src and (src.hate_per_powerful_hit or 0) > 0 and src.knowTalent and src:knowTalent(src.T_HATE_POOL) then
		local hateGain = 0
		local hateMessage

		if value / src.max_life > 0.33 then
			-- you deliver a big hit
			hateGain = hateGain + src.hate_per_powerful_hit
			hateMessage = "#F53CBE#Your powerful attack feeds your madness!"
		end

		if hateGain >= 0.1 then
			src.hate = math.min(src.max_hate, src.hate + hateGain)
			if hateMessage then
				game.logPlayer(src, hateMessage.." (+%d hate)", hateGain)
			end
		end
	end

	-- Bloodlust!
	if value > 0 and src and src.knowTalent and src:knowTalent(src.T_BLOODLUST) then
		src:setEffect(src.EFF_BLOODLUST, 1, {})
	end

	if value > 0 and self:knowTalent(self.T_RAMPAGE) then
		local t = self:getTalentFromId(self.T_RAMPAGE)
		t:onTakeHit(self, value / self.max_life)
	end

	-- Split ?
	if self.clone_on_hit and value >= self.clone_on_hit.min_dam_pct * self.max_life / 100 and rng.percent(self.clone_on_hit.chance) then
		-- Find space
		local x, y = util.findFreeGrid(self.x, self.y, 1, true, {[Map.ACTOR]=true})
		if x then
			-- Find a place around to clone
			local a
			if self.clone_base then a = self.clone_base:clone() else a = self:clone() end
			a.life = math.max(1, self.life - value / 2)
			a.clone_on_hit.chance = math.ceil(self.clone_on_hit.chance / 2)
			a.energy.value = 0
			a.exp_worth = 0.1
			a.inven = {}
			a:removeAllMOs()
			a.x, a.y = nil, nil
			game.zone:addEntity(game.level, a, "actor", x, y)
			game.logSeen(self, "%s splits in two!", self.name:capitalize())
			value = value / 2
		end
	end

	-- Apply Solipsism hit
	if damage_to_psi > 0 then
		local t = self:getTalentFromId(self.T_SOLIPSISM)
		local psi_damage_resist = 1 - t.getPsiDamageResist(self, t)/100
	--	print("Psi Damage Resist", psi_damage_resist, "Damage", damage_to_psi, "Final", damage_to_psi*psi_damage_resist)
		if self:getPsi() > damage_to_psi*psi_damage_resist then
			self:incPsi(-damage_to_psi*psi_damage_resist)
		else
			damage_to_psi = self:getPsi()
			self:incPsi(-damage_to_psi)
		end
		local mindcolor = DamageType:get(DamageType.MIND).text_color or "#aaaaaa#"
		game:delayedLogMessage(self, nil, "Solipsism hit", mindcolor.."#Source# converts some damage to Psi!")
		game:delayedLogDamage(src, self, damage_to_psi*psi_damage_resist, ("%s%d %s#LAST#"):format(mindcolor, damage_to_psi*psi_damage_resist, "to psi"), false)

		local oldval = value -- verbose

		value = value - damage_to_psi

		-- verbose addon start
		if self == game.player then
			print("$$$ Solipsism has converted damage of value: ", oldval - value, ", with this much damage remaining: ", value)
		end
		-- verbose addon end
	end

	if value <= 0 then return onTakeHit(self, 0, src, death_note) end
	-- VITALITY?
	if self:knowTalent(self.T_VITALITY) and self.life > self.max_life /2 and self.life - value <= self.max_life/2 then
		local t = self:getTalentFromId(self.T_VITALITY)
		t.do_vitality_recovery(self, t)
	end

	-- Daunting Presence?
	if value > (self.max_life / 20) and self:isTalentActive(self.T_DAUNTING_PRESENCE) then
		local t = self:getTalentFromId(self.T_DAUNTING_PRESENCE)
		t.do_daunting_presence(self, t)
	end

	-- Shield of Light
	tal = self:isTalentActive(self.T_SHIELD_OF_LIGHT)
	if tal then
		if value <= 2 then
			drain = value
		else
			drain = 2
		end
		if self:getPositive() >= drain then
			self:incPositive(- drain)

			-- Only calculate crit once per turn to avoid log spam
			if not self.turn_procs.shield_of_light_heal then
				local t = self:getTalentFromId(self.T_SHIELD_OF_LIGHT)
				self.turn_procs.shield_of_light_heal = true
				self.shield_of_light_heal = self:spellCrit(t.getHeal(self, t))
			end

			self:heal(self.shield_of_light_heal, tal)
		end
	end
	
	if self:attr("unstoppable") then
		if value > self.life - 1 then
			game:delayedLogDamage(src, self, 0, ("#RED#(%d refused)#LAST#"):format(value - (self.life - 1)), false)
			value = self.life - 1
			if self.life <= 1 then value = 0 end
			game:delayedLogMessage(self, nil, "unstoppable", "#RED##Source# is unstoppable!")
		end
	end

	if value >= self.life then
		local tal = self:isTalentActive(self.T_SECOND_LIFE)
		if tal then
			local sl = self:callTalent(self.T_SECOND_LIFE,"getLife")
			value = 0
			self.life = 1
			self:forceUseTalent(self.T_SECOND_LIFE, {ignore_energy=true})
			local value = self:heal(sl, self)
			game.logSeen(self, "#YELLOW#%s has been healed by a blast of positive energy!#LAST#", self.name:capitalize())
			if value > 0 then
				if self.player then world:gainAchievement("AVOID_DEATH", self) end
			end
		end

		local tal = self:isTalentActive(self.T_HEARTSTART)
		if tal then
			self:forceUseTalent(self.T_HEARTSTART, {ignore_energy=true})
		end
	end

	-- Shade's reform
	if value >= self.life and self.ai_state and self.ai_state.can_reform then
		local t = self:getTalentFromId(self.T_SHADOW_REFORM)
		if rng.percent(t.getChance(self, t)) then
			value = 0
			self.life = self.max_life
			game.logSeen(self, "%s fades for a moment and then reforms whole again!", self.name:capitalize())
			game.level.map:particleEmitter(self.x, self.y, 1, "teleport_out")
			game:playSoundNear(self, "talents/heal")
			game.level.map:particleEmitter(self.x, self.y, 1, "teleport_in")
		end
	end

	-- Shadow decoy
	if value >= self.life and self:isTalentActive(self.T_SHADOW_DECOY) then
		local t = self:getTalentFromId(self.T_SHADOW_DECOY)
		if t.onDie(self, t, value, src) then
			value = 0
		end
	end

	if value <= 0 then return onTakeHit(self, 0, src, death_note) end
	-- Vim leech
	if self:knowTalent(self.T_LEECH) and src.hasEffect and src:hasEffect(src.EFF_VIMSENSE) then
		local vt = self:getTalentFromId(self.T_LEECH)
		self:incVim(vt.getVim(self, vt))
		self:heal(vt.getHeal(self, vt), src)
		if self.player then src:logCombat(src, "#AQUAMARINE#You leech a part of #Target#'s vim.") end
	end

	-- Invisible on hit
	if value >= self.max_life * 0.10 and self:attr("invis_on_hit") and rng.percent(self:attr("invis_on_hit")) then
		self:setEffect(self.EFF_INVISIBILITY, 5, {power=self:attr("invis_on_hit_power")})
		for tid, _ in pairs(self.invis_on_hit_disable) do self:forceUseTalent(tid, {ignore_energy=true}) end
	end

	-- Bloodspring
	if value >= self.max_life * 0.15 and self:knowTalent(self.T_BLOODSPRING) then
		self:triggerTalent(self.T_BLOODSPRING)
	end

	if self:knowTalent(self.T_DUCK_AND_DODGE) then
		local t = self:getTalentFromId(self.T_DUCK_AND_DODGE)
		if value >= self.max_life * t.getThreshold(self, t) then
			self:setEffect(self.EFF_EVASION, t.getDuration(self, t), {chance=t.getEvasionChance(self, t), defense = t.getDefense(self)})
		end
	end

	-- Damage shield on hit
	if self:attr("contingency") and value >= self.max_life * self:attr("contingency") / 100 and not self:hasEffect(self.EFF_DAMAGE_SHIELD) then
		self:setEffect(self.EFF_DAMAGE_SHIELD, 3, {power=value * self:attr("contingency_shield") / 100})
		for tid, _ in pairs(self.contingency_disable) do self:forceUseTalent(tid, {ignore_energy=true}) end
	end

	-- Spell cooldowns on hit
	if self:attr("reduce_spell_cooldown_on_hit") and value >= self.max_life * self:attr("reduce_spell_cooldown_on_hit") / 100 then
		local alt = {}
		for tid, cd in pairs(self.talents_cd) do
			if rng.percent(self:attr("reduce_spell_cooldown_on_hit_chance")) then alt[tid] = true end
		end
		for tid, cd in pairs(alt) do
			self:alterTalentCoolingdown(tid, -1)
		end
	end

	-- Life leech
	if value > 0 and src and not src.dead and src.attr and src:attr("life_leech_chance") and rng.percent(src.life_leech_chance) then
		local leech = math.min(value, self.life) * src.life_leech_value / 100
		if leech > 0 then
			src:heal(leech, self)
			game:delayedLogMessage(src, self, "life_leech"..self.uid, "#CRIMSON##Source# leeches life from #Target#!")
		end
	end

	-- Life steal from weapon
	if value > 0 and src and not src.dead and src.attr and src:attr("lifesteal") then
		local leech = math.min(value, self.life) * src.lifesteal / 100
		if leech > 0 then
			src:heal(leech, self)
			game:delayedLogMessage(src, self, "lifesteal"..self.uid, "#CRIMSON##Source# steals life from #Target#!")
		end
	end

	-- Flat damage cap
	if self.flat_damage_cap and self.max_life and death_note and death_note.damtype then
		local oldval = value
		local cap = nil
		if self.flat_damage_cap.all then cap = self.flat_damage_cap.all end
		if self.flat_damage_cap[death_note.damtype] then cap = self.flat_damage_cap[death_note.damtype] end
		if cap and cap > 0 then
			local ignored = math.max(0, value - cap * self.max_life / 100)
			if ignored > 0 then 
				game:delayedLogDamage(src, self, 0, ("#LIGHT_GREY#(%d resilience)#LAST#"):format(ignored), false) 
				-- verbose addon start
				if self == game.player then
					if self:knowTalent(self.T_GRINDING_SHIELD) and self:isTalentActive(self.T_GRINDING_SHIELD) then
						print("$$$ Grinding Shield has ignored damage of value: ", ignored, ", with this much damage remaining: ", value - ignored)
					end
					if self:knowTalent(self.T_SUNCLOAK) and self:isTalentActive(self.T_SUNCLOAK) then
						print("$$$ Suncloak has ignored damage of value: ", ignored, ", with this much damage remaining: ", value - ignored)
					end
					print("$$$ flat_damage_cap has converted damage of value: ", ignored, ", with this much damage remaining: ", value - ignored)
				end
				-- verbose addon end
			end
			value = value - ignored
			-- print("[TAKE HIT] after flat damage cap", value) -- removed by Verbose, as it should only report when activated
		end
	end

	if self.on_takehit then value = self:check("on_takehit", value, src, death_note) end

	local eff = self:hasEffect(self.EFF_ELDRITCH_STONE)
	if eff then
		local abs = math.min(value, eff.power)
		self:incEquilibrium(abs * 2)
		if eff.power > abs then
			eff.power = eff.power - abs
			value = 0
		else
			value = value - abs
			self:removeEffect(self.EFF_ELDRITCH_STONE)
		end
		game:delayedLogDamage(src, self, 0, ("#SLATE#(%d to stone)#LAST#"):format(abs), false)
	end

	if self:knowTalent(self.T_STONESHIELD) and not self.turn_procs.stoneshield then
		local t = self:getTalentFromId(self.T_STONESHIELD)
		local m, mm, e, em = t.getValues(self, t)
		self:incMana(math.min(mm, value * m))
		self:incEquilibrium(-math.min(em, value * e))
		self.turn_procs.stoneshield = true
	end

	local eff = self:hasEffect(self.EFF_STONE_LINK)
	if eff then
		if eff.src:attr("dead") then
			self:removeEffect(self.EFF_STONE_LINK)
		else
			game:delayedLogMessage(eff.src, self, "stone_link"..(self.uid or ""), "#OLIVE_DRAB##Source# redirects damage from #Target# to %s!#LAST#", string.his_her_self(eff.src))
			game:delayedLogDamage(src, self, 0, ("#OLIVE_DRAB#(%d redirected)#LAST#"):format(value), false)
			eff.src:takeHit(value, src)
			game:delayedLogDamage(src, eff.src, value, ("#OLIVE_DRAB#%d redirected#LAST#"):format(value), false)
			value = 0
		end
	end

	local cb = {value=value}
	if self:fireTalentCheck("callbackOnHit", cb, src, death_note) then
		value = cb.value
	end

	local hd = {"Actor:takeHit", value=value, src=src, death_note=death_note}
	if self:triggerHook(hd) then value = hd.value end

	-- Resource leech
	if value > 0 and src and src.attr and src:attr("resource_leech_chance") and rng.percent(src.resource_leech_chance) then
		local leech = src.resource_leech_value
		src:incMana(leech)
		src:incVim(leech * 0.5)
		src:incPositive(leech * 0.25)
		src:incNegative(leech * 0.25)
		src:incEquilibrium(-leech * 0.35)
		src:incStamina(leech * 0.65)
		src:incHate(leech * 0.2)
		src:incPsi(leech * 0.2)
		game:delayedLogMessage(src, self, "resource_leech", "#CRIMSON##Source# leeches energies from #Target#!")
	end

	if self:knowTalent(self.T_DRACONIC_BODY) then
		local t = self:getTalentFromId(self.T_DRACONIC_BODY)
		t.trigger(self, t, value)
	end
	
	-- verbose addon start
	-- jx_value_end = jx_value_start - value
	if self == game.player then
		print("$$$ Total Reduction of damage from onTakeHit: ", jx_value_start - value)
	end
	-- verbose addon end

	return onTakeHit(self, value, src, death_note)
end

return _M